import asyncio
import inspect
import signal
from dataclasses import dataclass, field
from typing import Any, Callable

from volley.logging import logger


class GracefulKiller:
    def __init__(self) -> None:
        self.kill_now = False
        signal.signal(signal.SIGINT, self.exit_gracefully)
        signal.signal(signal.SIGTERM, self.exit_gracefully)

    def exit_gracefully(self, signum: int, frame: Any) -> None:  # pylint: disable=W0613
        logger.warning("Received kill: %s", signum)
        self.kill_now = True


@dataclass
class FuncEnvelope:
    func: Callable[..., Any]

    message_ctx_param: str = "msg_ctx"

    needs_msg_ctx: bool = field(init=False)
    is_coroutine: bool = field(init=False)
    message_param: str = field(init=False)

    def __post_init__(self) -> None:
        self.is_coroutine = asyncio.iscoroutinefunction(self.func)
        func_params = dict(inspect.signature(self.func).parameters)

        # validate parameters
        num_param_expected = 1
        self.needs_msg_ctx = self.message_ctx_param in func_params
        if self.needs_msg_ctx:
            num_param_expected = 2
            remaining_params = {k: v for k, v in func_params.items() if k != self.message_ctx_param}
        else:
            remaining_params = func_params
        # remaining param is name of the user provided message param
        if len(remaining_params) != 1:
            raise TypeError(
                f"Function '{self.func.__name__}' has too many arguments."
                f"Expected {num_param_expected}. Provided {func_params.keys()}"
            )

        # TODO: validation type annotation on message_param agrees with queue config
        self.message_param = list(remaining_params.keys())[0]



# Copyright (c) Shipt, Inc.
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import importlib
from pathlib import Path
from typing import Any, Dict, Optional, Type, Union

from pydantic import BaseModel
from yaml import Loader, load

from volley.connectors.base import BaseConsumer, BaseProducer
from volley.models.base import BaseModelHandler
from volley.serializers.base import BaseSerialization

GLOBALS = Path(__file__).parent.resolve().joinpath("global.yml")


def load_yaml(file_path: Union[str, Path]) -> Dict[str, Any]:
    """loads a yaml to dict from Path object
    Raises FileNotFoundError
    """
    if isinstance(file_path, str):
        path = Path(file_path)
    else:
        path = file_path
    with path.open() as f:
        cfg: Dict[str, Any] = load(f, Loader=Loader)
    return cfg


def import_module_from_string(module_str: str) -> type:
    """returns the module given its string path
    for example:
        'volley.data_models.GenericMessage'
    is equivalent to:
        from volley.data_models import GenericMessage
    """
    modules = module_str.split(".")
    class_obj = modules[-1]
    pathmodule = ".".join(modules[:-1])
    module = importlib.import_module(pathmodule)
    t: type = getattr(module, class_obj)
    return t


def get_configs() -> Dict[str, Dict[str, Any]]:
    return load_yaml(GLOBALS)


class QueueConfig(BaseModel):
    """Represents the configuration for a single queue. Provides the
    application with necessary configuration for interacting with the
    specified queue.

    Attributes:
        name (str): Alias for a particular queue.
        value (str): The system name for a queue.
            For example, the name of a Kafka topic (prd.my.long.kafka.topic.name) or name of a RSMQ queue.
        profile (Optional[str]): Either `kafka|rsmq`.
            Pertains to the type of connector required to produce and consume from the queue.
            If not provided, must provide values for ALL of `consumer, producer, serializer, model_handler`.
        data_model (Optional[str]): Defaults to `volley.data_models.GenericMessage`.
            Path to the Pydantic model used for data validation.
            When default is used, Volley will only validate that messages can be successfully
            converted to a Pydantic model or dictionary.
        serializer (Optional[str]): Defaults to `volley.serializers.OrJsonSerializer`.
            Path to the serializer.
        producer (Optional[str]): Used for providing a custom producer connector.
            Overrides the producer pertaining to that provided in `type`.
            Provide the dot path to the producer class. e.g. for Kafka,defaults to
            `volley.connectors.kafka.KafkaProducer`;
            cf. [Extending Connectors](./connectors/connectors.md#extending-connectors-with-plugins).
        consumer (Optional[str]): Used for providing a custom consumer connector.
            Overrides the consumer pertaining to that provided in `type`.
            Provide the dot path to the consumer class.
            e.g. for Kafka, defaults to `volley.connectors.kafka.KafkaConsumer`;
            cf. [Extending Connectors](./connectors/connectors.md#extending-connectors-with-plugins).
        config: (Optional[str]): Any configuration to be passed directly to the queue connector. For example,
            all [librdkafka configurations](https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md)
            can be passed through to the connector via a dictionary here.

    # Example

    ```python
    from volley import Engine, QueueConfig

    input_cfg = QueueConfig(
        name="my-input-queue",
        value="my.input.kafka.topic",
        profile="confluent",
        config={"consumer.group": "my_consumer_group", "bootstrap.servers": "kafka:9092"}
    )
    output_cfg = QueueConfig(
        name="my-output-queue",
        value="my.output.kafka.topic",
        profile="confluent",
        config={"bootstrap.servers": "kafka:9092"}
    )
    app = Engine(
        app_name="my-app",
        input_queue="my-input-queue",
        output_queues=["my-output-queue"],
        queue_config=[input_cfg, output_cfg]
    )
    ```
    """

    name: str
    value: str
    profile: Optional[str] = None
    consumer: Optional[Union[str, Type[BaseConsumer]]] = None
    producer: Optional[Union[str, Type[BaseProducer]]] = None
    data_model: Optional[Union[str, Any]] = None
    model_handler: Optional[Union[str, Type[BaseModelHandler]]] = None
    serializer: Optional[Union[str, Type[BaseSerialization]]] = None
    config: Optional[Dict[Any, Any]] = None

    def to_dict(self) -> Dict[str, Any]:
        """transform to dictionary omitting optional fields when not provided"""
        return {k: v for k, v in self.__dict__.items() if v is not None}



import asyncio
import contextvars
import functools
from typing import Any, Awaitable, Callable

from prometheus_client import Counter

from volley.util import FuncEnvelope

APP_STATUS = Counter("volley_app_completions", "Application cycle status", ["app_name", "status"])  # success/failure


async def run_in_threadpool(func: Callable[..., Any], *args: Any) -> Any:
    loop = asyncio.get_event_loop()
    child = functools.partial(func, *args)
    context = contextvars.copy_context()
    func = context.run
    args = (child,)
    return await loop.run_in_executor(None, func, *args)


async def run_worker_function(
    f: FuncEnvelope,
    message: Any,
    ctx: Any,
    app_name: str = "volley",
) -> Any:
    """Handles executing the function provides by user.
    `async def` functions need to get called in threadpool.
    If user requests message context in their function, then volley needs to provide
    it to the wrapped function as well.

    Args:
        f: Envelope passed from volley.Engine
        message: the message constructed by user specified message and model handler
        ctx: the raw consumed object from queue, can contain `message`
        app_name: Defined an volley.Engine init. Defaults to "volley"

    Returns:
        Whatever is returned by the user's python function
    """
    if f.needs_msg_ctx:
        f.func = functools.partial(f.func, **{f.message_ctx_param: ctx})
    try:
        if f.is_coroutine:
            fun_result = await f.func(message)
        else:
            fun_result = await run_in_threadpool(f.func, message)
    except Exception:
        APP_STATUS.labels(app_name=app_name, status="failure").inc()
        raise
    APP_STATUS.labels(app_name=app_name, status="success").inc()
    return fun_result


def run_async(func: Callable[..., Awaitable[Any]]) -> Callable[..., None]:
    def wrapper() -> None:
        asyncio.run(func())

    return wrapper



import logging
import os
import signal
from dataclasses import dataclass, field
from threading import Thread
from typing import Any, Dict, List, Optional, Union

from confluent_kafka import Consumer, Message, Producer
from prometheus_client import Counter

from volley.connectors.base import BaseConsumer, BaseProducer
from volley.data_models import QueueMessage

logger = logging.getLogger(__name__)

DELIVERY_STATUS = Counter("delivery_report_status", "Kafka delivered message", ["status"])


@dataclass
class ConfluentKafkaConsumer(BaseConsumer):
    """
    Use for consuming a single message from a single Kafka topic.

    Built on confluent-kafka-python/librdkafka. Offsets are stored in librdkafka
        and committed according to auto.commit.interval.ms
        https://github.com/edenhill/librdkafka/blob/master/CONFIGURATION.md

    ## Multi-topic subscription
    You may find yourself wanting to configure a single Volley worker to consume from multiple Kafka topics,
    each with the same message schema. You can do this by specifying the queue value as a comma
    separated list of topics:

    ```python hl_lines="4"
    cfg = {
        "input-topic": {
            "profile": "confluent",
            "value": "kafka.topic.0,kafka.topic.1",
            "consumer": "volley.connectors.confluent.ConfluentKafkaConsumer",
            "config": {"bootstrap.servers": "kafka:9092", "group.id": "myConsumerGroup"},
        },
    }
    ```

    Note: multi-topic is only supported for consumption.
    Volley will only be able to *consume* to the queue as configured above.
    To produce to multiple topics, specify each topic as a separate queue and publish to them.

    ```python
    return [
        ("output-topic-0", message_object),
        ("output-topic-1", message_object)
    ]
    ```

    """

    poll_interval: float = 10
    auto_offset_reset: str = "earliest"
    auto_commit_interval_ms: int = 3000
    stop_on_failure: bool = True
    # mapping of topic -> partition/offset of the last stored commit
    last_offset: Dict[str, Dict[int, int]] = field(init=False)

    def __post_init__(self) -> None:  # noqa: C901
        self.config = _handle_creds(self.config, True)

        if "auto.offset.reset" not in self.config:
            logger.info("Assigning auto.offset.reset default: %s", self.auto_offset_reset)
            self.config["auto.offset.reset"] = self.auto_offset_reset

        if "auto.commit.interval.ms" not in self.config:
            logger.info("Assigning auto.commit.interval.ms default: %s", self.auto_commit_interval_ms)
            self.config["auto.commit.interval.ms"] = self.auto_commit_interval_ms

        if "stop_on_failure" in self.config:
            self.stop_on_failure = self.config["stop_on_failure"]

        if self.stop_on_failure:
            logger.info(
                "`stop_on_failure` = %s. Application will gracefully shutdown when" "there are downstream failures.",
                self.stop_on_failure,
            )
        else:
            logger.warning(
                "`stop_on_failure` = %s. Application will log a CRITICAL level message"
                "and continue processing on downstream failures.",
                self.stop_on_failure,
            )
        # self.config provided from base Consumer class
        # consumer group assignment
        # try config, then env var, then command line argument w/ env
        if "group.id" in self.config:
            # we'll pass the config directly into Kafka constructor
            pass
        else:
            try:
                self.config["group.id"] = os.environ["KAFKA_CONSUMER_GROUP"]
            except KeyError:
                logger.exception("KAFKA_CONSUMER_GROUP not specified in env variable or group.id in config dict")
                raise

        if "poll_interval" in self.config:
            # poll_interval can be overridden by a user provided Engine init config
            # but it should not be passed to base Confluent Consumer init
            # confluent will ignore it with a warning, however
            self.poll_interval = self.config.pop("poll_interval")

        # this is not overridable, due to the behavior in self.on_success()
        # on_success calls store_offsets()
        self.config["enable.auto.offset.store"] = False
        self.c = Consumer(self.config, logger=logger)

        # https://github.com/apache/kafka/blob/
        # 8007211cc982d8458223e866c1ee7d94b69e0249/core/src/main/scala/kafka/common/Topic.scala#L29
        # "," are not allowed in Apache Kafka topic names.
        # Allow multi-topic subscription by supplying a comma separated lists of topics
        topics: List[str] = self.queue_name.split(",")
        self.c.subscribe(topics)

        self.last_offset = {topic: {} for topic in topics}

        logger.info("Subscribed to %s", topics)

    def consume(  # type: ignore
        self,
    ) -> Optional[QueueMessage]:
        message = self.c.poll(self.poll_interval)
        if message is None:
            pass
        elif message.error():
            logger.warning(message.error())
            message = None
        else:
            return QueueMessage(message_context=message, message=message.value())

    def on_success(self, message_context: Message) -> None:
        """stores any offsets that are able to be stored"""
        topic = message_context.topic()
        partition = message_context.partition()
        this_offset = message_context.offset()

        # see if we've already committed this or a higher offset
        # delivery report from kafka producer are not guaranteed to be in order
        # that they were produced
        # https://github.com/confluentinc/confluent-kafka-python/issues/300#issuecomment-358416432

        try:
            last_commit = self.last_offset[topic][partition]
        except KeyError:
            # first message from this topic-partition
            self.last_offset[topic][partition] = this_offset
            self.c.store_offsets(message_context)
            return
        if this_offset > last_commit:
            self.c.store_offsets(message_context)  # committed according to auto.commit.interval.ms
            self.last_offset[topic][partition] = this_offset

    def on_fail(self, message_context: Message) -> None:
        logger.critical(
            "Downstream failure. Did not commit topic: %s, partition: %d, offset: %d, message: %s.",
            message_context.topic(),
            message_context.partition(),
            message_context.offset(),
            message_context.value(),
        )
        if self.stop_on_failure:
            logger.critical("Downstream failure. Stopping application.")
            # explicitly call the connectors shutdown
            self.shutdown()
            # this send a kill command to py-volley engine to handle any other graceful shutdown procedures
            os.kill(os.getpid(), signal.SIGINT)

    def shutdown(self) -> None:
        self.c.close()
        logger.info("Successfully commit offsets and left consumer group %s", self.config.get("group.id"))


@dataclass
class ConfluentKafkaProducer(BaseProducer):
    compression_type: str = "gzip"
    thread: bool = False
    poll_thread_timeout: float = 1.0

    def __post_init__(self) -> None:  # noqa: C901
        self.callback_delivery = True
        self.config = _handle_creds(self.config, False)
        if "compression.type" not in self.config:
            logger.info("Assigning compression.type default: %s", self.compression_type)
            self.config["compression.type"] = self.compression_type

        self.p = Producer(self.config, logger=logger)

        # producer poll thread
        self.kill_poll_thread = False
        if "poll_thread_timeout" in self.config:
            self.poll_thread_timeout = self.config.pop("poll_thread_timeout")

    def init_callbacks(self, consumer: BaseConsumer, thread: bool = True) -> None:
        # daemon thread - this can be aggressively killed on shutdown
        # Volley will call shutdown() which will trigger confluent.Producer.flush()
        # flush() will trigger acked() and call consumers on_success/on_fail
        self.on_success = consumer.on_success
        self.on_fail = consumer.on_fail
        if thread:
            self.thread = True
            self.poll_thread = Thread(target=self.handle_poll, daemon=True)
            self.poll_thread.start()

    def handle_poll(self) -> None:
        while not self.kill_poll_thread:
            self.p.poll(0.2)

    def acked(self, err: Optional[str], msg: Message, consumer_context: Any) -> None:
        if err is not None:
            logger.critical(
                "Failed delivery to %s, message: %s, error: %s",
                msg.topic(),
                msg.value(),
                err,
            )
            DELIVERY_STATUS.labels("fail").inc()
            self.on_fail(consumer_context)  # type: ignore

        else:
            logger.debug(
                "Successful delivery to %s, partion: %d, offset: %d", msg.topic(), msg.partition(), msg.offset()
            )
            DELIVERY_STATUS.labels("success").inc()
            self.on_success(consumer_context)  # type: ignore

    def produce(self, queue_name: str, message: bytes, message_context: Any, **kwargs: Union[str, int]) -> bool:
        self.p.produce(
            key=kwargs.get("key"),
            topic=queue_name,
            value=message,
            headers=kwargs.get("headers"),
            callback=lambda err, msg, obj=message_context: self.acked(err, msg, obj),
        ),

        logger.debug("Sent to topic: %s", queue_name)
        return True

    def shutdown(self) -> None:
        self.p.flush()
        if self.thread:
            self.kill_poll_thread = True
            self.poll_thread.join(self.poll_thread_timeout)


def _handle_creds(config_dict: Dict[str, Any], is_consumer: bool) -> Dict[str, Any]:
    """Populates credential related kafka configuration values based on well
    known environment variables.

    Environment variable mappings:
        * KAFKA_BROKERS maps to bootstrap.servers
        * KAFKA_CONSUMER_BROKERS maps to bootstrap.servers for consumers only,
          falling back to KAFKA_BROKERS.
        * KAFKA_PRODUCER_BROKERS maps to bootstrap.servers for producers only,
          falling back to KAFKA_BROKERS.
        * KAFKA_KEY maps to sasl.username
        * KAFKA_SECRET maps to sasl.password

    Note that environment variables will only be used when previous config
    values do not exist.
    """

    if "bootstrap.servers" not in config_dict:
        # Attempt to set consumer / producer specific overrides first
        if is_consumer and os.getenv("KAFKA_CONSUMER_BROKERS") is not None:
            config_dict["bootstrap.servers"] = os.environ["KAFKA_CONSUMER_BROKERS"]

        if not is_consumer and os.getenv("KAFKA_PRODUCER_BROKERS") is not None:
            config_dict["bootstrap.servers"] = os.environ["KAFKA_PRODUCER_BROKERS"]

    # Fall back to "KAFKA_BROKERS" for both consumers and producers if still not set
    if "bootstrap.servers" not in config_dict:
        if os.getenv("KAFKA_BROKERS") is not None:
            config_dict["bootstrap.servers"] = os.environ["KAFKA_BROKERS"]

    if "bootstrap.servers" not in config_dict:
        logger.error("Kafka brokers not specified, set in config dict or env var KAFKA_BROKERS")
        raise ValueError("bootstrap.servers is required")

    # No key == dev mode
    if "sasl.username" not in config_dict and "sasl.passsword" not in config_dict:
        sasl_username = os.getenv("KAFKA_KEY")
        sasl_password = os.getenv("KAFKA_SECRET")
        if (sasl_username is not None) and (sasl_password is not None):
            logger.info(
                "KAFKA_KEY and KAFKA_SECRET found in environment. Assigning security.protocol and sasl.mechanism"
            )
            config_dict["sasl.username"] = sasl_username
            config_dict["sasl.password"] = sasl_password
            config_dict["security.protocol"] = "SASL_SSL"
            config_dict["sasl.mechanism"] = "PLAIN"
    return config_dict


class BatchJsonConfluentConsumer(ConfluentKafkaConsumer):
    """Consumes multiple messages before process.
    Parses messages to byte string json

    ## Configuration
    You may find yourself wanting to configure a single Volley worker to consume
        and process multiple messages per cycle.
    For example, consume 10 messages for a batch write to a Redis cache.

    Use the batch Kafka Consumer by either specifying `profile`:

    ```python hl_lines="3"
    cfg = {
        "input-topic": {
            "profile": "confluent-batch-json",
            "value": "kafka.topic.0,kafka.topic.1",
            "config": {
                "bootstrap.servers": "kafka:9092",
                "group.id": "myConsumerGroup",
                "batch_size" 10,
                "batch_time_seconds": 2.5,
            },
        },
    }
    ```

    Or by overriding the consumer:

    ```python hl_lines="5"
    cfg = {
        "input-topic": {
            "value": "kafka.topic.0,kafka.topic.1",
            "profile": "confluent",
            "consumer": "volley.connectors.confluent.ConfluentKafkaConsumer",
            "config": {
                "bootstrap.servers": "kafka:9092",
                "group.id": "myConsumerGroup",
                "batch_size" 10,
                "batch_time_seconds": 2.5,
            },
        },
    }
    ```
    Your application then receives a `List` of 10 messages (assuming there were 10 available to be consumed).

    """

    batch_size: int = 5
    batch_time_seconds: float = 2.0

    def __post_init__(self) -> None:
        # handle some configurations before setting ConfluentKafkaConsumer
        if "batch_size" in self.config:
            self.batch_size = self.config.pop("batch_size")
        if "batch_time_seconds" in self.config:
            self.batch_time_seconds = self.config.pop("batch_time_seconds")
        super().__post_init__()

    def consume(
        self,
    ) -> Optional[QueueMessage]:
        """consumes self.config["batch_size"] messages from the input topic
        if self.config["batch_time_seconds"] reached, return however many messages currently consumed
        other wise keep polling for messages until batch_size is reached
        """

        messages: List[Optional[bytes]] = []
        raw_messages: List[Union[Message, None]] = self.c.consume(
            num_messages=self.batch_size, timeout=self.batch_time_seconds
        )
        for_commit: List[Message] = []
        if not len(raw_messages):
            logger.debug("No messages")
            return None
        for m in raw_messages:
            if m is None:
                continue
            elif m.error():
                logger.error(m.error())
                continue
            else:
                for_commit.append(m)
                messages.append(m.value())
        num_messages = len(messages)
        if num_messages:
            logger.debug("messages in batch/batch_size %s / %s", num_messages, self.batch_size)
            all_msg: bytes = b"[" + b",".join(messages) + b"]"  # type: ignore
            return QueueMessage(message_context=for_commit, message=all_msg)
        else:
            return None

    def on_success(self, message_context: List[Message]) -> None:
        """stores any offsets that are able to be stored"""

        # see if we've already committed this or a higher offset
        # delivery report from kafka producer are not guaranteed to be in order
        # that they were produced
        # https://github.com/confluentinc/confluent-kafka-python/issues/300#issuecomment-358416432
        for m in message_context:
            topic = m.topic()
            partition = m.partition()
            this_offset = m.offset()

            try:
                last_commit = self.last_offset[topic][partition]
            except KeyError:
                # first message from this topic-partition
                self.last_offset[topic][partition] = this_offset
                self.c.store_offsets(m)
                continue
            if this_offset > last_commit:
                self.c.store_offsets(m)  # committed according to auto.commit.interval.ms
                self.last_offset[topic][partition] = this_offset



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['volley', 'volley.connectors', 'volley.models', 'volley.serializers']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.4.1,<6.1',
 'msgpack>=1.0.3,<2.0.0',
 'orjson>=3.6.4,<4.0',
 'prometheus-client>=0.11.0,<1',
 'pydantic>=1.8.2,<2',
 'starlette<1',
 'uvicorn<1']

extras_require = \
{'all': ['confluent-kafka>=1.8.2,<2.0.0',
         'PyRSMQ>=0.4.5,<0.5.0',
         'hiredis>=2.0,<3.0',
         'tenacity>=8.0.1,<9.0.0',
         'pyzmq>=22.3.0'],
 'kafka': ['confluent-kafka>=1.8.2,<2.0.0'],
 'rsmq': ['PyRSMQ>=0.4.5,<0.5.0',
          'hiredis>=2.0,<3.0',
          'tenacity>=8.0.1,<9.0.0'],
 'zmq': ['pyzmq>=22.3.0']}

setup_kwargs = {
    'name': 'py-volley',
    'version': '1.0.0',
    'description': 'Pluggable message queueing for Python',
    'long_description': '![VolleyFull-Horizontal](https://user-images.githubusercontent.com/81711984/149005139-f0441dcf-c76e-4112-baf1-998d0a6abdbb.png)\n\nDocumentation: https://shipt.github.io/py-volley/\n\nVolley makes building event stream applications easier and more accessible. Use Volley if you need to quickly develop an application to consume messages, processes them (and do other things), then publish results to one or many places. Volley was inspired ease of use and developer experience provided by the [Flask](https://github.com/pallets/flask) and [FastAPI](https://github.com/tiangolo/fastapi) projects, and aims to make working with queue based and event driven system as accessible as REST APIs.\n\nVolley handles a number of operations that need to happen before and after processing a message. Reading the data, serialization, data validation, all need to happen before data reaches your application. If these fail, Volley can route the message to a dead-letter-queue. After processing, Volley again handles data validation, serialization, and the writing/publishing of data to any number of output queues. Finally, upon successfully delivery of that message to the target queue, Volley handles marking it as read or deleting it from the input queue.\n\nAll of Volley\'s major operations (connectors, serializers, data validation/model handling) can be extended with plugins, and comes with built in support for queues-like technologies [Apache Kafka](https://kafka.apache.org/) and [RSMQ](https://github.com/mlasevich/PyRSMQ) (Redis Simple Message Queue). There is a plugin built for a Postgres queue in our [examples](./example/plugins/my_plugin.py).\n\n\n[![Build Status](https://drone.shipt.com/api/badges/shipt/py-volley/status.svg?ref=refs/heads/main)](https://drone.shipt.com/shipt/py-volley)\n[![Coverage](https://sonarqube.shipt.com/api/project_badges/measure?project=shipt_py-volley_AYImTs5MsYUjTdFQ7Awt&metric=coverage&token=squ_e98968a6b1bce0281e001fd0e70e538f6228b47f)](https://sonarqube.shipt.com/dashboard?id=shipt_py-volley_AYImTs5MsYUjTdFQ7Awt)\n\n# Installation\n\nRequires Python >= 3.8\n\n```bash\npip install py-volley[all]\n```\n\nYou can also limit the dependencies by:\n```bash\npip install py-volley[kafka]  # Kafka dependencies\npip install py-volley[rsmq]  # RSMQ dependencies\npip install py-volley[zmq]  # ZeroMQ dependencies\n```\n\n## Features\n- Built in support for [Apache Kafka](https://kafka.apache.org/), [RSMQ](https://github.com/smrchy/rsmq), [ZeroMQ](https://zeromq.org/)\n- [Prometheus](https://prometheus.io/) metrics for all operations such as function processing time, and consumption and production count.\n- Serialization in JSON and [MessagePack](https://msgpack.org/index.html)\n- Data validation via [Pydantic](https://pydantic-docs.helpmanual.io/)\n- Optionally configured integration with dead-letter-queues\n- Extendible connectors (consumer/producers), serializers, model handlers, and model handlers via plugins.\n\n## Getting started\n\nVolley handles the process of consuming/producing by providing developers with extendible interfaces and handlers:\n- connectors - consumer and producer interfaces which define how the application should read messages, write messages, and what actions to take when a message is successfully or fails processing.\n- serializers - handlers and interface which describe the behavior for reading an byte objects from connectors. For example, Json or MessagePack serializers.\n- model_handler - handler and interface which works very closely with serializers. Typically used to turn serialized data into a structured Python data model. Pydantic is Volley\'s most supported data_model and can handler serialization itself.\n- data_model - When your application receives data from a queue, what schema and object do you expect it in? The data_model is provided by the user. And the `model_handler` describes how to construct your `data_model`.\n\n\nTo demonstrate, let\'s create an application with two worker nodes. One consumes from Kafka, finds the maximum value in a list then publishes it to Redis. The other consumes the message from Redis - if the max value is > 10, it logs to console otherwise it constructs a new list and publishes to the same Kafka topic. \n\n```mermaid\nflowchart LR\nA[(Kafka)] --> |consume| B[Worker 1]\nB --> |publish| C[(Redis)]\nC --> |consume| D[Worker 2]\nD --> E{>10}\nE --> | no | A\nE --> | yes | F[Log to Console]\n```\n\nYou can skip the details and just run `make intro.start`, which runs this example through `./example/intro/docker-compose.yml`\n\n1. start Kafka and Redis instance\n\n```\ndocker run -d -p 6379:6379 redis:5.0.0\ndocker run -d -p 9092:9092 bashj79/kafka-kraft\n```\n\n2. Configure the queues and data models. Let\'s put this in `./my_config.py`.\n\n```python\n# ./my_config.py\nfrom typing import List, Tuple\nfrom pydantic import BaseModel\n\nfrom volley import Engine, QueueConfig\n\n# define the schemas for the first and second worker nodes.\nclass InputMessage(BaseModel):\n  my_values: List[float]\n\nclass OutputMessage(BaseModel):\n  the_max: float\n\n# define the configurations for the two queues, one in Kafka and the other in Redis.\nqueue_config = [\n  QueueConfig(\n    name="my-kafka-input",\n    value="my.kafka.topic.name",\n    profile="confluent",\n    data_model=InputMessage,\n    config={\n      "group.id": "my.consumer.group",\n      "bootstrap.servers": "localhost:9092",\n    }\n  ),\n  QueueConfig(\n    name="my-redis-output",\n    value="my.redis.output.queue.name",\n    profile="rsmq",\n    data_model=OutputMessage,\n    config={\n      "host": "localhost",\n      "port": 6379,\n    }\n  )\n]\n```\n\n3. Build the first worker node - consume from Kafka, find the max value, publish to Redis\n\n```python\n# ./app_0.py\nfrom typing import List, Tuple\nfrom volley import Engine\n\nfrom my_config import queue_config, InputMessage, OutputMessage\n# the first node - reads from kafka and writes to redis\napp_0 = Engine(\n  app_name="app_0",\n  input_queue="my-kafka-input",  # one input\n  output_queues=["my-redis-output"],  # zero to many outputs\n  queue_config=queue_config\n)\n\n@app_0.stream_app\ndef kafka_to_redis(msg: InputMessage) -> List[Tuple[str, OutputMessage]]:\n  print(f"Received {msg.json()}")\n  max_val = max(msg.my_values)\n  out = OutputMessage(the_max=max_val)\n  print(out)\n  return [("my-redis-output", out)]  # a list of one or many output targets and messages\n\nif __name__ == "__main__":\n  kafka_to_redis()\n\n```\n\n4. Run the first application in a terminal\n```bash\npython app_0.py\n```\n\n\n5. Build the second worker node - consume from Redis, determine if we log to console or recycle the message as a new list.\n```python\n# ./app_1.py\nfrom typing import List, Tuple, Union\nfrom volley import Engine\n\nfrom my_config import OutputMessage, queue_config, InputMessage\n\n# the second node\napp_1 = Engine(\n  app_name="app_1",\n  input_queue="my-redis-output",\n  output_queues=["my-kafka-input"],\n  queue_config=queue_config,\n  metrics_port=None\n)\n\n@app_1.stream_app\ndef redis_to_kafka(msg: OutputMessage) -> Union[bool, List[Tuple[str, InputMessage]]]:\n  print(f"The maximum: {msg.the_max}")\n  if msg.the_max > 10:\n    print("That\'s it, we are done!")\n    return True\n  else:\n    out = InputMessage(my_values=[msg.the_max, msg.the_max+1, msg.the_max+2])\n    return [("my-kafka-input", out)]  # a list of one or many output targets and messages\n\nif __name__ == "__main__":\n    redis_to_kafka()\n```\n\n6. Run the second worker node in another terminal\n```bash\npython app_1.py\n```\n\n7. Finally, let\'s manually publish a message to the input kafka topic:\n```python\nfrom confluent_kafka import Producer\nimport json\nproducer = Producer({"bootstrap.servers": "localhost:9092"})\nproducer.produce(topic="my.kafka.topic.name", value=json.dumps({"my_values":[1,2,3]}))\nproducer.flush(5)\n```\n\nYou should see the following in your two terminals\n\n__./app_0.py__\n```\nReceived {"my_values": [1.0, 2.0, 3.0]}\nthe_max=3.0\nReceived {"my_values": [3.0, 4.0, 5.0]}\nthe_max=5.0\nReceived {"my_values": [5.0, 6.0, 7.0]}\nthe_max=7.0\nReceived {"my_values": [7.0, 8.0, 9.0]}\nthe_max=9.0\nReceived {"my_values": [9.0, 10.0, 11.0]}\nthe_max=11.0\n```\n\n__./app_1.py__\n```\nThe maximum: 3.0\nThe maximum: 5.0\nThe maximum: 7.0\nThe maximum: 9.0\nThe maximum: 11.0\nThat\'s it, we are done!\n```\n\n# Complete example\n\nClone this repo and `make run.example` to see a complete example of:\n- consuming a message from a Kafka topic\n- producing to RSMQ\n- consuming from RSMQ and publishing to Kafka and Postgres using custom plugin for Postgres.\n\n# Contributing\n\nSee our [contributing guide](./CONTRIBUTING.md).\n\nThanks goes to great [projects](./ATTRIBUTIONS.md) and these incredible people.\n\n<a href="https://github.com/shipt/py-volley/graphs/contributors">\n  <img src="https://contrib.rocks/image?repo=shipt/py-volley" />\n</a>\n',
    'author': 'ask-machine-learning',
    'author_email': 'shipt@shipt.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<4',
}


setup(**setup_kwargs)



import os
import threading

import uvicorn
from prometheus_client import (
    CONTENT_TYPE_LATEST,
    CollectorRegistry,
    generate_latest,
    multiprocess,
    start_http_server,
)
from starlette.applications import Starlette
from starlette.requests import Request
from starlette.responses import Response
from starlette.routing import Route

from volley.logging import logger


def multiproc_collector_server() -> Starlette:
    """Serves a multiprocess collector
    https://github.com/prometheus/client_python#multiprocess-mode-eg-gunicorn
    """

    async def app(request: Request) -> Response:  # pylint: disable=W0613
        prometheus_registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(prometheus_registry)
        data = generate_latest(prometheus_registry)
        headers = {
            "Content-type": CONTENT_TYPE_LATEST,
            "Content-length": str(len(data)),
        }
        return Response(data, media_type=CONTENT_TYPE_LATEST, headers=headers)

    return Starlette(
        routes=[
            Route("/metrics", app),
        ]
    )


def serve_metrics(port: int) -> None:
    """Serves the built-in prometheus webserver by default.
    Serves multiprocess collector when PROMETHEUS_MULTIPROC_DIR present in env vars.

    Both methods run the webserver on a daemon thread
    """
    if os.getenv("PROMETHEUS_MULTIPROC_DIR") is not None:
        logger.info("Serving multi-process collector")
        server = multiproc_collector_server()
        t = threading.Thread(target=uvicorn.run, args=(server,), kwargs={"host": "0.0.0.0", "port": port}, daemon=True)
        t.start()
    else:
        start_http_server(port=port)
